/*
 * This file is part of SpongeAPI, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.api.world.gen.populators;

import org.spongepowered.api.world.gen.Populator;

/**
 * Represents a populator which places in a pseudo path of ice.
 */
public interface IcePath extends Populator {

    /**
     * Gets the radius of the path to spawn.
     * 
     * @return The path radius
     */
    int getRadius();

    /**
     * Sets the radius of the path to spawn, cannot be negative.
     * 
     * @param radius The new path radius
     */
    void setRadius(int radius);

    /**
     * Gets the number of sections to attempt to spawn per chunk, must be
     * greater than zero.
     * 
     * @return The number to spawn
     */
    int getSectionsPerChunk();

    /**
     * Sets the number of sections to attempt to spawn per chunk, must be
     * greater than zero.
     * 
     * @param sections The new amount to spawn
     */
    void setSectionsPerChunk(int sections);

    /**
     * A builder for constructing {@link IcePath} populators.
     */
    interface Builder {

        /**
         * Sets the radius of the path to spawn, cannot be negative.
         * 
         * @param radius The new path radius
         * @return This builder, for chaining
         */
        Builder radius(int radius);

        /**
         * Sets the number of sections to attempt to spawn per chunk, must be
         * greater than zero.
         * 
         * @param sections The new amount to spawn
         * @return This builder, for chaining
         */
        Builder perChunk(int sections);

        /**
         * Resets this builder to the default values.
         * 
         * @return This builder, for chaining
         */
        Builder reset();

        /**
         * Builds a new instance of a {@link IcePath} populator with the
         * settings set within the builder.
         * 
         * @return A new instance of the populator
         * @throws IllegalStateException If there are any settings left unset
         *             which do not have default values
         */
        IcePath build() throws IllegalStateException;

    }

}
